from mongoengine import *

from spaceone.core.model.mongo_model import MongoModel


class PlannedLimit(EmbeddedDocument):
    date = StringField(required=True, max_length=7)
    limit = FloatField(default=0)


class ProviderFilter(EmbeddedDocument):
    state = StringField(max_length=20, default='ENABLED', choices=('ENABLED', 'DISABLED'))
    providers = ListField(StringField(), default=[])

    def to_dict(self):
        return dict(self.to_mongo())


class Notification(EmbeddedDocument):
    threshold = FloatField(required=True)
    unit = StringField(max_length=20, required=True, choices=('PERCENT', 'ACTUAL_COST'))
    notification_type = StringField(max_length=20, required=True, choices=('CRITICAL', 'WARNING'))
    notified_months = ListField(StringField(), default=[])

    def to_dict(self):
        return dict(self.to_mongo())


class Budget(MongoModel):
    budget_id = StringField(max_length=40, generate_id='budget', unique=True)
    name = StringField(max_length=255, default='')
    limit = FloatField(required=True)
    planned_limits = ListField(EmbeddedDocumentField(PlannedLimit), default=[])
    currency = StringField()
    provider_filter = EmbeddedDocumentField(ProviderFilter, required=True)
    time_unit = StringField(max_length=20, choices=('TOTAL', 'MONTHLY'))
    start = StringField(required=True, max_length=7)
    end = StringField(required=True, max_length=7)
    notifications = ListField(EmbeddedDocumentField(Notification), default=[])
    tags = DictField(default={})
    project_id = StringField(max_length=40, default=None, null=True)
    project_group_id = StringField(max_length=40, default=None, null=True)
    data_source_id = StringField(max_length=40)
    domain_id = StringField(max_length=40)
    created_at = DateTimeField(auto_now_add=True)
    updated_at = DateTimeField(auto_now=True)

    meta = {
        'updatable_fields': [
            'name',
            'limit',
            'planned_limits',
            'provider_filter',
            'notifications',
            'tags'
        ],
        'minimal_fields': [
            'budget_id',
            'name',
            'limit',
            'provider_filter',
            'project_id',
            'project_group_id',
            'data_source_id'
        ],
        'change_query_keys': {
            'user_projects': 'project_id',
            'user_project_groups': 'project_group_id'
        },
        'ordering': ['name'],
        'indexes': [
            'name',
            'project_id',
            'project_group_id',
            'data_source_id',
            'domain_id'
        ]
    }



import logging

from spaceone.core.manager import BaseManager
from spaceone.cost_analysis.manager.plugin_manager import PluginManager
from spaceone.cost_analysis.manager.data_source_rule_manager import DataSourceRuleManager
from spaceone.cost_analysis.service.data_source_rule_service import DataSourceRuleService
from spaceone.cost_analysis.connector.datasource_plugin_connector import DataSourcePluginConnector
from spaceone.cost_analysis.model.data_source_model import DataSource

_LOGGER = logging.getLogger(__name__)


class DataSourcePluginManager(BaseManager):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.dsp_connector: DataSourcePluginConnector = self.locator.get_connector('DataSourcePluginConnector')

    def initialize(self, endpoint):
        _LOGGER.debug(f'[initialize] data source plugin endpoint: {endpoint}')
        self.dsp_connector.initialize(endpoint)

    def init_plugin(self, options, domain_id):
        plugin_info = self.dsp_connector.init(options, domain_id)

        _LOGGER.debug(f'[plugin_info] {plugin_info}')
        plugin_metadata = plugin_info.get('metadata', {})

        return plugin_metadata

    def verify_plugin(self, options, secret_data, schema, domain_id):
        self.dsp_connector.verify(options, secret_data, schema, domain_id)

    def get_tasks(self, options, secret_id, secret_data, schema, start, last_synchronized_at, domain_id):
        response = self.dsp_connector.get_tasks(options, secret_data, schema, domain_id, start, last_synchronized_at)
        tasks = response.get('tasks', [])

        for task in tasks:
            task.update({
                'secret_id': secret_id,
                'schema': schema
            })

        return tasks, response.get('changed', [])

    def get_cost_data(self, options, secret_data, schema, task_options, domain_id):
        return self.dsp_connector.get_cost_data(options, secret_data, schema, task_options, domain_id)

    def get_data_source_plugin_endpoint_by_vo(self, data_source_vo: DataSource):
        plugin_info = data_source_vo.plugin_info.to_dict()
        endpoint, updated_version = self.get_data_source_plugin_endpoint(plugin_info, data_source_vo.domain_id)

        if updated_version:
            _LOGGER.debug(
                f'[get_data_source_plugin_endpoint_by_vo] upgrade plugin version: {plugin_info["version"]} -> {updated_version}')
            self.upgrade_data_source_plugin_version(data_source_vo, endpoint, updated_version)

        return endpoint

    def get_data_source_plugin_endpoint(self, plugin_info, domain_id):
        plugin_mgr: PluginManager = self.locator.get_manager('PluginManager')
        return plugin_mgr.get_plugin_endpoint(plugin_info, domain_id)

    def upgrade_data_source_plugin_version(self, data_source_vo: DataSource, endpoint, updated_version):
        plugin_info = data_source_vo.plugin_info.to_dict()

        data_source_id = data_source_vo.data_source_id
        domain_id = data_source_vo.domain_id

        self.initialize(endpoint)

        plugin_options = plugin_info.get('options', {})

        plugin_metadata = self.init_plugin(plugin_options, domain_id)
        plugin_info['version'] = updated_version
        plugin_info['metadata'] = plugin_metadata
        data_source_vo.update({'plugin_info': plugin_info})

        self.delete_data_source_rules(data_source_id, domain_id)
        self.create_data_source_rules_by_metadata(plugin_metadata, data_source_id, domain_id)

    def delete_data_source_rules(self, data_source_id, domain_id):
        _LOGGER.debug(f'[_delete_data_source_rules] delete all data source rules: {data_source_id}')
        data_source_rule_mgr: DataSourceRuleManager = self.locator.get_manager('DataSourceRuleManager')
        old_data_source_rule_vos = data_source_rule_mgr.filter_data_source_rules(data_source_id=data_source_id,
                                                                                 rule_type='MANAGED',
                                                                                 domain_id=domain_id)

        old_data_source_rule_vos.delete()

    def create_data_source_rules_by_metadata(self, metadata, data_source_id, domain_id):
        data_source_rules = metadata.get('data_source_rules', [])

        if len(data_source_rules) > 0:
            _LOGGER.debug(f'[_create_data_source_rules_by_metadata] create data source rules: {data_source_id} / '
                          f'rule count = {len(data_source_rules)}')

            data_source_rule_svc: DataSourceRuleService = self.locator.get_service('DataSourceRuleService', metadata)
            for data_source_rule_params in data_source_rules:
                data_source_rule_params['data_source_id'] = data_source_id
                data_source_rule_params['domain_id'] = domain_id
                data_source_rule_params['rule_type'] = 'MANAGED'
                data_source_rule_svc.create_data_source_rule(data_source_rule_params)



import logging

from spaceone.core.manager import BaseManager
from spaceone.cost_analysis.model.data_source_model import DataSource

_LOGGER = logging.getLogger(__name__)


class DataSourceManager(BaseManager):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.data_source_model: DataSource = self.locator.get_model('DataSource')

    def register_data_source(self, params):
        def _rollback(data_source_vo):
            _LOGGER.info(f'[register_data_source._rollback] '
                         f'Delete data source : {data_source_vo.name} '
                         f'({data_source_vo.data_source_id})')
            data_source_vo.delete()

        data_source_vo: DataSource = self.data_source_model.create(params)
        self.transaction.add_rollback(_rollback, data_source_vo)

        return data_source_vo

    def update_data_source(self, params):
        data_source_vo: DataSource = self.get_data_source(params['data_source_id'], params['domain_id'])
        return self.update_data_source_by_vo(params, data_source_vo)

    def update_data_source_by_vo(self, params, data_source_vo):
        def _rollback(old_data):
            _LOGGER.info(f'[update_data_source_by_vo._rollback] Revert Data : '
                         f'{old_data["data_source_id"]}')
            data_source_vo.update(old_data)

        self.transaction.add_rollback(_rollback, data_source_vo.to_dict())
        return data_source_vo.update(params)

    def deregister_data_source(self, data_source_id, domain_id):
        data_source_vo: DataSource = self.get_data_source(data_source_id, domain_id)
        data_source_vo.delete()

    @staticmethod
    def deregister_data_source_by_vo(data_source_vo):
        data_source_vo.delete()

    def get_data_source(self, data_source_id, domain_id, only=None):
        return self.data_source_model.get(data_source_id=data_source_id, domain_id=domain_id, only=only)

    def filter_data_sources(self, **conditions):
        return self.data_source_model.filter(**conditions)

    def list_data_sources(self, query={}):
        return self.data_source_model.query(**query)

    def stat_data_sources(self, query):
        return self.data_source_model.stat(**query)



import functools
from spaceone.api.cost_analysis.v1 import cost_pb2
from spaceone.core.pygrpc.message_type import *
from spaceone.core import utils
from spaceone.cost_analysis.model.cost_model import Cost

__all__ = ['CostInfo', 'CostsInfo']


def CostInfo(cost_vo: Cost, minimal=False):
    info = {
        'cost_id': cost_vo.cost_id,
        'cost': cost_vo.cost,
        'provider': cost_vo.provider,
        'region_code': cost_vo.region_code,
        'product': cost_vo.product,
        'usage_type': cost_vo.usage_type,
        'resource': cost_vo.resource,
        'data_source_id': cost_vo.data_source_id,
        'billed_date': cost_vo.billed_date
    }

    if not minimal:
        info.update({
            'usage_quantity': cost_vo.usage_quantity,
            'usage_unit': cost_vo.usage_unit,
            'tags': change_struct_type(cost_vo.tags),
            'additional_info': change_struct_type(cost_vo.additional_info),
            'service_account_id': cost_vo.service_account_id,
            'project_id': cost_vo.project_id,
            'project_group_id': cost_vo.project_group_id,
            'data_source_id': cost_vo.data_source_id,
            'domain_id': cost_vo.domain_id,
            'billed_year': cost_vo.billed_year,
            'billed_month': cost_vo.billed_month,
        })

    return cost_pb2.CostInfo(**info)


def CostsInfo(cost_vos, total_count, **kwargs):
    return cost_pb2.CostsInfo(results=list(
        map(functools.partial(CostInfo, **kwargs), cost_vos)), total_count=total_count)



import logging
import functools

from spaceone.core import utils
from spaceone.core.manager import BaseManager
from spaceone.cost_analysis.manager.identity_manager import IdentityManager
from spaceone.cost_analysis.model.data_source_rule_model import DataSourceRule, DataSourceRuleCondition

_LOGGER = logging.getLogger(__name__)


class DataSourceRuleManager(BaseManager):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.data_source_rule_model: DataSourceRule = self.locator.get_model('DataSourceRule')
        self.identity_mgr: IdentityManager = self.locator.get_manager('IdentityManager')
        self._project_info = {}
        self._service_account_info = {}
        self._data_source_rule_info = {}

    def create_data_source_rule(self, params):
        def _rollback(data_source_rule_vo: DataSourceRule):
            _LOGGER.info(f'[create_data_source_rule._rollback] '
                         f'Delete event rule : {data_source_rule_vo.name} '
                         f'({data_source_rule_vo.data_source_rule_id})')
            data_source_rule_vo.delete()

        data_source_rule_vo: DataSourceRule = self.data_source_rule_model.create(params)
        self.transaction.add_rollback(_rollback, data_source_rule_vo)

        return data_source_rule_vo

    def update_data_source_rule(self, params):
        data_source_rule_vo: DataSourceRule = self.get_data_source_rule(params['data_source_rule_id'],
                                                                        params['domain_id'])
        return self.update_data_source_rule_by_vo(params, data_source_rule_vo)

    def update_data_source_rule_by_vo(self, params, data_source_rule_vo):
        def _rollback(old_data):
            _LOGGER.info(f'[update_data_source_rule_by_vo._rollback] Revert Data : '
                         f'{old_data["data_source_rule_id"]}')
            data_source_rule_vo.update(old_data)

        self.transaction.add_rollback(_rollback, data_source_rule_vo.to_dict())

        return data_source_rule_vo.update(params)

    def delete_data_source_rule(self, data_source_rule_id, domain_id):
        data_source_rule_vo: DataSourceRule = self.get_data_source_rule(data_source_rule_id, domain_id)
        self.delete_data_source_rule_by_vo(data_source_rule_vo)

    def delete_data_source_rule_by_vo(self, data_source_rule_vo):
        data_source_rule_vo.delete()

    def get_data_source_rule(self, data_source_rule_id, domain_id, only=None):
        return self.data_source_rule_model.get(data_source_rule_id=data_source_rule_id, domain_id=domain_id, only=only)

    def filter_data_source_rules(self, **conditions):
        return self.data_source_rule_model.filter(**conditions)

    def list_data_source_rules(self, query={}):
        return self.data_source_rule_model.query(**query)

    def stat_data_source_rules(self, query):
        return self.data_source_rule_model.stat(**query)

    def change_cost_data(self, cost_data):
        data_source_id = cost_data['data_source_id']
        domain_id = cost_data['domain_id']
        managed_data_source_rule_vos, custom_data_source_rule_vos = self._get_data_source_rules(data_source_id, domain_id)

        cost_data = self._apply_data_source_rule_to_cost_data(cost_data, managed_data_source_rule_vos, domain_id)
        cost_data = self._apply_data_source_rule_to_cost_data(cost_data, custom_data_source_rule_vos, domain_id)

        return cost_data

    def _apply_data_source_rule_to_cost_data(self, cost_data, data_source_rule_vos, domain_id):
        for data_source_rule_vo in data_source_rule_vos:
            is_match = self._change_cost_data_by_rule(cost_data, data_source_rule_vo)
            if is_match:
                cost_data = self._change_cost_data_with_actions(cost_data, data_source_rule_vo.actions, domain_id)

            if is_match and data_source_rule_vo.options.stop_processing:
                break

        return cost_data

    def _change_cost_data_with_actions(self, cost_data, actions, domain_id):
        for action, value in actions.items():
            if action == 'change_project':
                cost_data['project_id'] = value

            elif action == 'match_project':
                source = value['source']
                target_key = value.get('target', 'project_id')
                target_value = utils.get_dict_value(cost_data, source)
                if target_value:
                    project_info = self._get_project(target_key, target_value, domain_id)
                    if project_info:
                        cost_data['project_id'] = project_info['project_id']

            elif action == 'match_service_account':
                source = value['source']
                target_key = value.get('target', 'service_account_id')
                target_value = utils.get_dict_value(cost_data, source)
                if target_value:
                    service_account_info = self._get_service_account(target_key, target_value, domain_id)
                    if service_account_info:
                        cost_data['service_account_id'] = service_account_info['service_account_id']
                        cost_data['project_id'] = service_account_info.get('project_info', {}).get('project_id')

            if action == 'add_additional_info':
                cost_data['additional_info'] = cost_data.get('additional_info', {})
                cost_data['additional_info'].update(value)

        return cost_data

    def _get_service_account(self, target_key, target_value, domain_id):
        if f'service-account:{domain_id}:{target_key}:{target_value}' in self._service_account_info:
            return self._service_account_info[f'service-account:{domain_id}:{target_key}:{target_value}']

        query = {
            'filter': [
                {'k': target_key, 'v': target_value, 'o': 'eq'},
                {'k': 'service_account_type', 'v': 'GENERAL', 'o': 'eq'}
            ],
            'only': ['service_account_id', 'project_info']
        }

        response = self.identity_mgr.list_service_accounts(query, domain_id)
        results = response.get('results', [])
        total_count = response.get('total_count', 0)

        service_account_info = None
        if total_count > 0:
            service_account_info = results[0]

        self._service_account_info[f'service-account:{domain_id}:{target_key}:{target_value}'] = service_account_info
        return service_account_info

    def _get_project(self, target_key, target_value, domain_id):
        if f'project:{domain_id}:{target_key}:{target_value}' in self._project_info:
            return self._project_info[f'project:{domain_id}:{target_key}:{target_value}']

        query = {
            'filter': [
                {'k': target_key, 'v': target_value, 'o': 'eq'}
            ],
            'only': ['project_id']
        }

        response = self.identity_mgr.list_projects(query, domain_id)
        results = response.get('results', [])
        total_count = response.get('total_count', 0)

        project_info = None
        if total_count > 0:
            project_info = results[0]

        self._project_info[f'project:{domain_id}:{target_key}:{target_value}'] = project_info
        return project_info

    def _change_cost_data_by_rule(self, cost_data, data_source_rule_vo: DataSourceRule):
        conditions_policy = data_source_rule_vo.conditions_policy

        if conditions_policy == 'ALWAYS':
            return True
        else:
            results = list(map(functools.partial(self._check_condition, cost_data), data_source_rule_vo.conditions))

            if conditions_policy == 'ALL':
                return all(results)
            else:
                return any(results)

    @staticmethod
    def _check_condition(cost_data, condition: DataSourceRuleCondition):
        cost_value = utils.get_dict_value(cost_data, condition.key)
        condition_value = condition.value
        operator = condition.operator

        if cost_value is None:
            return False

        if operator == 'eq':
            if cost_value == condition_value:
                return True
            else:
                return False
        elif operator == 'contain':
            if cost_value.lower().find(condition_value.lower()) >= 0:
                return True
            else:
                return False
        elif operator == 'not':
            if cost_value != condition_value:
                return True
            else:
                return False
        elif operator == 'not_contain':
            if cost_value.lower().find(condition_value.lower()) < 0:
                return True
            else:
                return False

        return False

    def _get_data_source_rules(self, data_source_id, domain_id):
        if data_source_id in self._data_source_rule_info:
            return self._data_source_rule_info[data_source_id].get('managed', []), \
                   self._data_source_rule_info[data_source_id].get('custom', [])

        managed_query = self._make_data_source_rule_query(data_source_id, 'MANAGED', domain_id)
        managed_data_source_rule_vos, total_count = self.list_data_source_rules(managed_query)

        custom_query = self._make_data_source_rule_query(data_source_id, 'CUSTOM', domain_id)
        custom_data_source_rule_vos, total_count = self.list_data_source_rules(custom_query)

        self._data_source_rule_info[data_source_id] = {}
        self._data_source_rule_info[data_source_id]['managed'] = managed_data_source_rule_vos
        self._data_source_rule_info[data_source_id]['custom'] = custom_data_source_rule_vos

        return managed_data_source_rule_vos, custom_data_source_rule_vos

    @staticmethod
    def _make_data_source_rule_query(data_source_id, rule_type, domain_id):
        return {
            'filter': [
                {'k': 'data_source_id', 'v': data_source_id, 'o': 'eq'},
                {'k': 'domain_id', 'v': domain_id, 'o': 'eq'},
                {'k': 'rule_type', 'v': rule_type, 'o': 'eq'}
            ],
            'sort': {'key': 'order'}
        }



import logging

from spaceone.core import config
from spaceone.core.manager import BaseManager
from spaceone.core.connector.space_connector import SpaceConnector
from spaceone.core import utils
from spaceone.cost_analysis.error import *

_LOGGER = logging.getLogger(__name__)


class SecretManager(BaseManager):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.secret_connector: SpaceConnector = self.locator.get_connector('SpaceConnector',
                                                                           service='secret',
                                                                           token=config.get_global('TOKEN'))

    def create_secret(self, domain_id, secret_data, schema):
        def _rollback(secret_id):
            _LOGGER.info(f'[create_secret._rollback] Delete secret : {secret_id}')
            self.delete_secret(secret_id, domain_id)

        params = {
            'name': utils.generate_id('secret-cost-data-source'),
            'data': secret_data,
            'secret_type': 'CREDENTIALS',
            'schema': schema,
            'domain_id': domain_id
        }

        response = self.secret_connector.dispatch('Secret.create', params)
        _LOGGER.debug(f'[_create_secret] {response}')
        secret_id = response['secret_id']

        self.transaction.add_rollback(_rollback, secret_id)

        return secret_id

    def delete_secret(self, secret_id, domain_id):
        params = {
            'secret_id': secret_id,
            'domain_id': domain_id
        }
        self.secret_connector.dispatch('Secret.delete', params)

    def list_secrets(self, query, domain_id):
        return self.secret_connector.dispatch('Secret.list', {'query': query, 'domain_id': domain_id})

    def get_secret(self, secret_id, domain_id):
        return self.secret_connector.dispatch('Secret.get', {'secret_id': secret_id, 'domain_id': domain_id})

    def get_secret_data(self, secret_id, domain_id):
        response = self.secret_connector.dispatch('Secret.get_data', {'secret_id': secret_id, 'domain_id': domain_id})
        return response['data']



import logging

from spaceone.core.service import *
from spaceone.core import utils
from spaceone.cost_analysis.error import *
from spaceone.cost_analysis.manager.cost_manager import CostManager
from spaceone.cost_analysis.manager.identity_manager import IdentityManager
from spaceone.cost_analysis.model.cost_model import Cost

_LOGGER = logging.getLogger(__name__)


@authentication_handler
@authorization_handler
@mutation_handler
@event_handler
class CostService(BaseService):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.cost_mgr: CostManager = self.locator.get_manager('CostManager')

    @transaction(append_meta={'authorization.scope': 'PROJECT'})
    @check_required(['cost', 'data_source_id', 'billed_date', 'domain_id'])
    def create(self, params):
        """Register cost

        Args:
            params (dict): {
                'cost': 'float',
                'usage_quantity': 'float',
                'usage_unit': 'str',
                'provider': 'str',
                'region_code': 'str',
                'product': 'str',
                'usage_type': 'str',
                'resource': 'str',
                'tags': 'dict',
                'additional_info': 'dict',
                'service_account_id': 'str',
                'project_id': 'str',
                'data_source_id': 'str',
                'billed_date': 'str',
                'domain_id': 'str'
            }

        Returns:
            cost_vo (object)
        """

        # validation check (service_account_id / project_id / data_source_id)

        cost_vo: Cost = self.cost_mgr.create_cost(params)

        self.cost_mgr.remove_stat_cache(params['domain_id'], params['data_source_id'])

        return cost_vo

    @transaction(append_meta={'authorization.scope': 'PROJECT'})
    @check_required(['cost_id', 'domain_id'])
    def delete(self, params):
        """Deregister cost

        Args:
            params (dict): {
                'cost_id': 'str',
                'domain_id': 'str'
            }

        Returns:
            None
        """

        domain_id = params['domain_id']

        cost_vo: Cost = self.cost_mgr.get_cost(params['cost_id'], params['domain_id'])

        self.cost_mgr.remove_stat_cache(domain_id, cost_vo.data_source_id)

        self.cost_mgr.delete_cost_by_vo(cost_vo)

    @transaction(append_meta={'authorization.scope': 'PROJECT'})
    @check_required(['cost_id', 'domain_id'])
    def get(self, params):
        """ Get cost

        Args:
            params (dict): {
                'cost_id': 'str',
                'domain_id': 'str',
                'only': 'list
            }

        Returns:
            cost_vo (object)
        """

        cost_id = params['cost_id']
        domain_id = params['domain_id']

        return self.cost_mgr.get_cost(cost_id, domain_id, params.get('only'))

    @transaction(append_meta={'authorization.scope': 'PROJECT'})
    @check_required(['data_source_id', 'domain_id'])
    @append_query_filter(['cost_id', 'provider', 'region_code', 'region_key', 'product', 'usage_type', 'resource',
                          'service_account_id', 'project_id', 'project_group_id', 'data_source_id', 'domain_id',
                          'user_projects'])
    @append_keyword_filter(['cost_id'])
    @set_query_page_limit(1000)
    def list(self, params):
        """ List costs

        Args:
            params (dict): {
                'cost_id': 'str',
                'provider': 'str',
                'region_code': 'str',
                'region_key': 'str',
                'product': 'str',
                'usage_type': 'str',
                'resource': 'str',
                'service_account_id': 'str',
                'project_id': 'str',
                'project_group_id': 'str',
                'data_source_id': 'str'
                'domain_id': 'str',
                'query': 'dict (spaceone.api.core.v1.Query)',
                'user_projects': 'list' // from meta
            }

        Returns:
            cost_vos (object)
            total_count
        """

        query = params.get('query', {})
        return self.cost_mgr.list_costs(query)

    @transaction(append_meta={'authorization.scope': 'PROJECT'})
    @check_required(['query', 'query.granularity', 'query.start', 'query.end', 'query.fields', 'data_source_id',
                     'domain_id'])
    @append_query_filter(['data_source_id', 'domain_id', 'user_projects'])
    @append_keyword_filter(['cost_id'])
    @set_query_page_limit(1000)
    def analyze(self, params):
        """
        Args:
            params (dict): {
                'query': 'dict (spaceone.api.core.v1.TimeSeriesAnalyzeQuery)',
                'data_source_id': 'str',
                'domain_id': 'str',
                'user_projects': 'list' // from meta
            }

        Returns:
            values (list) : 'list of statistics data'

        """

        domain_id = params['domain_id']
        data_source_id = params['data_source_id']
        query = params.get('query', {})

        return self.cost_mgr.analyze_costs_by_granularity(query, domain_id, data_source_id)

    @transaction(append_meta={'authorization.scope': 'PROJECT'})
    @check_required(['query', 'domain_id'])
    @append_query_filter(['data_source_id', 'domain_id', 'user_projects'])
    @append_keyword_filter(['cost_id'])
    @set_query_page_limit(1000)
    def stat(self, params):
        """
        Args:
            params (dict): {
                'query': 'dict (spaceone.api.core.v1.StatisticsQuery)',
                'data_source_id': 'str',
                'domain_id': 'str',
                'user_projects': 'list' // from meta
            }

        Returns:
            values (list) : 'list of statistics data'

        """

        domain_id = params['domain_id']
        query = params.get('query', {})

        if self._is_distinct_query(query):
            page, query = self._get_page_from_query(query)
            search, query = self._get_search_value_from_query(query)
            query_hash = utils.dict_to_hash(query)

            self.cost_mgr.create_cost_query_history(query, query_hash, domain_id, 'global')

            response = self.cost_mgr.stat_monthly_costs_with_cache(query, query_hash, domain_id, 'global')

            if search:
                response = self._search_results(response, search)

            if page:
                response = self._page_results(response, page)

            return response
        else:
            raise ERROR_NOT_SUPPORT_QUERY_OPTION(query_option='aggregate')

    @staticmethod
    def _is_distinct_query(query):
        if 'distinct' in query:
            return True
        else:
            return False

    @staticmethod
    def _get_page_from_query(query):
        if 'page' in query:
            page = query['page']
            del query['page']
        else:
            page = None

        return page, query

    @staticmethod
    def _get_search_value_from_query(query):
        distinct = query['distinct']

        search = None
        changed_filter = []
        for condition in query.get('filter', []):
            key = condition.get('key', condition.get('k'))
            value = condition.get('value', condition.get('v'))
            operator = condition.get('operator', condition.get('o'))

            if key == distinct and operator == 'contain':
                search = value
            else:
                changed_filter.append(condition)

        query['filter'] = changed_filter

        return search, query

    @staticmethod
    def _search_results(response, search):
        search = search.lower()
        changed_results = []

        for result in response.get('results', []):
            if search in result.lower():
                changed_results.append(result)

        return {
            'results': changed_results,
        }

    @staticmethod
    def _page_results(response, page):
        results = response.get('results', [])
        response = {
            'total_count': len(results)
        }

        if 'limit' in page and page['limit'] > 0:
            start = page.get('start', 1)
            if start < 1:
                start = 1

            response['results'] = results[start - 1:start + page['limit'] - 1]
        else:
            response['results'] = results

        return response



import logging
import copy
from typing import List
from datetime import datetime, timedelta

from spaceone.core.error import *
from spaceone.core import queue, utils, config
from spaceone.core.manager import BaseManager
from spaceone.cost_analysis.model.job_model import Job
from spaceone.cost_analysis.model.cost_model import CostQueryHistory
from spaceone.cost_analysis.manager.cost_manager import CostManager

_LOGGER = logging.getLogger(__name__)


class JobManager(BaseManager):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.job_model: Job = self.locator.get_model('Job')
        self.job_timeout = config.get_global('JOB_TIMEOUT', 7200)
        self.cost_mgr: CostManager = self.locator.get_manager('CostManager')

    def is_job_running(self, data_source_id, domain_id):
        job_vos: List[Job] = self.job_model.filter(data_source_id=data_source_id, domain_id=domain_id,
                                                   status='IN_PROGRESS')

        running_job_count = job_vos.count()

        for job_vo in job_vos:
            if datetime.utcnow() > (job_vo.created_at + timedelta(seconds=self.job_timeout)):
                self.change_timeout_status(job_vo)
                running_job_count -= 1

        return running_job_count > 0

    def create_job(self, data_source_id, domain_id, job_options, total_tasks, changed=None):
        job_options['no_preload_cache'] = job_options.get('no_preload_cache', False)
        job_options['start'] = job_options.get('start')

        data = {
            'data_source_id': data_source_id,
            'domain_id': domain_id,
            'options': job_options
        }

        if total_tasks:
            data.update({
                'total_tasks': total_tasks,
                'remained_tasks': total_tasks
            })

        if changed:
            data['changed'] = changed

        job_vo = self.job_model.create(data)

        _LOGGER.debug(f'[create_job] create job: {job_vo.job_id}')
        return job_vo

    def update_job_by_vo(self, params, job_vo):
        return job_vo.update(params)

    def get_job(self, job_id, domain_id, only=None):
        return self.job_model.get(job_id=job_id, domain_id=domain_id, only=only)

    def filter_jobs(self, **conditions):
        return self.job_model.filter(**conditions)

    def list_jobs(self, query={}):
        return self.job_model.query(**query)

    def stat_jobs(self, query):
        return self.job_model.stat(**query)

    def preload_cost_stat_queries(self, domain_id, data_source_id):
        cost_query_cache_time = config.get_global('COST_QUERY_CACHE_TIME', 3)
        cache_time = datetime.utcnow() - timedelta(days=cost_query_cache_time)

        query = {
            'filter': [
                {'k': 'domain_id', 'v': domain_id, 'o': 'eq'},
                {'k': 'data_source_id', 'v': data_source_id, 'o': 'eq'},
                {'k': 'updated_at', 'v': cache_time, 'o': 'gte'},
            ]
        }

        _LOGGER.debug(f'[_preload_cost_stat_queries] cost_query_cache_time: {cost_query_cache_time} days')

        history_vos, total_count = self.cost_mgr.list_cost_query_history(query)
        for history_vo in history_vos:
            _LOGGER.debug(f'[_preload_cost_stat_queries] create query cache: {history_vo.query_hash}')
            self._create_cache_by_history(history_vo, domain_id)

    def _create_cache_by_history(self, history_vo: CostQueryHistory, domain_id):
        query = history_vo.query_options
        query_hash = history_vo.query_hash
        data_source_id = history_vo.data_source_id

        # Original Date Range
        self._create_cache(copy.deepcopy(query), query_hash, domain_id, data_source_id)

    def _create_cache(self, query, query_hash, domain_id, data_source_id):
        if granularity := query.get('granularity'):
            if granularity == 'DAILY':
                self.cost_mgr.analyze_costs_with_cache(query, query_hash, domain_id, data_source_id)
            elif granularity == 'MONTHLY':
                self.cost_mgr.analyze_monthly_costs_with_cache(query, query_hash, domain_id, data_source_id)
            elif granularity == 'YEARLY':
                self.cost_mgr.analyze_yearly_costs_with_cache(query, query_hash, domain_id, data_source_id)
        else:
            self.cost_mgr.stat_monthly_costs_with_cache(query, query_hash, domain_id, data_source_id)

    @staticmethod
    def decrease_remained_tasks(job_vo: Job):
        return job_vo.decrement('remained_tasks', 1)

    @staticmethod
    def change_success_status(job_vo: Job):
        _LOGGER.info(f'[change_success_status] job success: {job_vo.job_id}')

        return job_vo.update({
            'status': 'SUCCESS',
            'finished_at': datetime.utcnow()
        })

    @staticmethod
    def change_canceled_status(job_vo: Job):
        _LOGGER.error(f'[change_canceled_status], job canceled ({job_vo.job_id})')

        return job_vo.update({
            'status': 'CANCELED',
            'finished_at': datetime.utcnow()
        })

    @staticmethod
    def change_timeout_status(job_vo: Job):
        _LOGGER.error(f'[change_timeout_status] job timeout: {job_vo.job_id}')

        return job_vo.update({
            'status': 'TIMEOUT',
            'finished_at': datetime.utcnow()
        })

    @staticmethod
    def change_error_status(job_vo: Job, e):
        if not isinstance(e, ERROR_BASE):
            e = ERROR_UNKNOWN(message=str(e))

        _LOGGER.error(f'[change_error_status] job error ({job_vo.job_id}): {e.message}', exc_info=True)

        job_vo.update({
            'status': 'FAILURE',
            'error_code': e.error_code,
            'error_message': e.message,
            'finished_at': datetime.utcnow()
        })



import logging

from spaceone.core.service import *
from spaceone.cost_analysis.error import *
from spaceone.cost_analysis.manager.budget_usage_manager import BudgetUsageManager

_LOGGER = logging.getLogger(__name__)


@authentication_handler
@authorization_handler
@mutation_handler
@event_handler
class BudgetUsageService(BaseService):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.budget_usage_mgr: BudgetUsageManager = self.locator.get_manager('BudgetUsageManager')

    @transaction(append_meta={'authorization.scope': 'PROJECT'})
    @check_required(['domain_id'])
    @append_query_filter(['budget_id', 'data_source_id', 'name', 'date', 'domain_id'])
    @append_keyword_filter(['budget_id', 'name'])
    @set_query_page_limit(1000)
    def list(self, params):
        """ List budget_usages

        Args:
            params (dict): {
                'budget_id': 'str',
                'data_source_id': 'str',
                'name': 'str',
                'date': 'str',
                'domain_id': 'str',
                'query': 'dict (spaceone.api.core.v1.Query)',
                'user_projects': 'list', // from meta,
                'user_project_groups': 'list', // from meta
            }

        Returns:
            budget_usage_vos (object)
            total_count
        """

        query = self._set_user_project_or_project_group_filter(params)
        return self.budget_usage_mgr.list_budget_usages(query)

    @transaction(append_meta={'authorization.scope': 'PROJECT'})
    @check_required(['query', 'domain_id'])
    @append_query_filter(['data_source_id', 'domain_id'])
    @append_keyword_filter(['budget_id', 'name'])
    @set_query_page_limit(1000)
    def stat(self, params):
        """
        Args:
            params (dict): {
                'data_source_id': 'str',
                'domain_id': 'str',
                'query': 'dict (spaceone.api.core.v1.StatisticsQuery)',
                'user_projects': 'list', // from meta,
                'user_project_groups': 'list' // from meta
            }

        Returns:
            values (list) : 'list of statistics data'

        """

        query = self._set_user_project_or_project_group_filter(params)
        return self.budget_usage_mgr.stat_budget_usages(query)

    @transaction(append_meta={'authorization.scope': 'PROJECT'})
    @check_required(['query', 'query.fields', 'domain_id'])
    @append_query_filter(['data_source_id', 'domain_id'])
    @append_keyword_filter(['budget_id', 'name'])
    @set_query_page_limit(1000)
    def analyze(self, params):
        """
        Args:
            params (dict): {
                'query': 'dict (spaceone.api.core.v1.TimeSeriesAnalyzeQuery)',
                'domain_id': 'str',
                'user_projects': 'list', // from meta
                'user_project_groups': 'list' // from meta
            }

        Returns:
            values (list) : 'list of statistics data'

        """

        query = self._set_user_project_or_project_group_filter(params)
        self._check_granularity(query.get('granularity'))

        return self.budget_usage_mgr.analyze_budget_usages(query)

    @staticmethod
    def _check_granularity(granularity):
        if granularity and granularity != 'MONTHLY':
            raise ERROR_INVALID_PARAMETER(key='query.granularity',
                                          reason='Granularity is only MONTHLY.')

    @staticmethod
    def _set_user_project_or_project_group_filter(params):
        query = params.get('query', {})
        query['filter'] = query.get('filter', [])

        if 'user_projects' in params:
            user_projects = params['user_projects'] + [None]
            query['filter'].append({'k': 'user_projects', 'v': user_projects, 'o': 'in'})

        if 'user_project_groups' in params:
            user_project_groups = params['user_project_groups'] + [None]
            query['filter'].append({'k': 'user_project_groups', 'v': user_project_groups, 'o': 'in'})

        return query



from mongoengine import *

from spaceone.core.model.mongo_model import MongoModel


class ProviderFilter(EmbeddedDocument):
    state = StringField(max_length=20, default='ENABLED', choices=('ENABLED', 'DISABLED'))
    providers = ListField(StringField(), default=[])

    def to_dict(self):
        return dict(self.to_mongo())


class BudgetUsage(MongoModel):
    budget_id = StringField(max_length=40, required=True)
    name = StringField(max_length=255, default='')
    date = StringField(max_length=7, required=True)
    cost = FloatField(required=True)
    limit = FloatField(required=True)
    currency = StringField(default=None, null=True)
    provider_filter = EmbeddedDocumentField(ProviderFilter, required=True)
    budget = ReferenceField('Budget', reverse_delete_rule=CASCADE)
    project_id = StringField(max_length=40, default=None, null=True)
    project_group_id = StringField(max_length=40, default=None, null=True)
    data_source_id = StringField(max_length=40)
    domain_id = StringField(max_length=40)
    updated_at = DateTimeField(auto_now=True)

    meta = {
        'updatable_fields': [
            'name',
            'cost',
            'limit'
        ],
        'minimal_fields': [
            'budget_id',
            'name',
            'date',
            'usd_cost',
            'limit'
        ],
        'change_query_keys': {
            'user_projects': 'project_id',
            'user_project_groups': 'project_group_id'
        },
        'ordering': ['budget_id', 'date'],
        'indexes': [
            'budget_id',
            'name',
            'date',
            'project_id',
            'project_group_id',
            'data_source_id',
            'domain_id'
        ]
    }



import logging

from spaceone.core import config
from spaceone.core.manager import BaseManager
from spaceone.core.connector.space_connector import SpaceConnector

_LOGGER = logging.getLogger(__name__)


class NotificationManager(BaseManager):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.notification_connector: SpaceConnector = self.locator.get_connector('SpaceConnector',
                                                                                 service='notification',
                                                                                 token=config.get_global('TOKEN'))

    def create_notification(self, message):
        _LOGGER.debug(f'Notify message: {message}')
        return self.notification_connector.dispatch('Notification.create', message)



import logging
from datetime import datetime
from dateutil.rrule import rrule, MONTHLY

from spaceone.core.manager import BaseManager
from spaceone.core import utils
from spaceone.cost_analysis.manager.identity_manager import IdentityManager
from spaceone.cost_analysis.manager.notification_manager import NotificationManager
from spaceone.cost_analysis.manager.cost_manager import CostManager
from spaceone.cost_analysis.manager.budget_manager import BudgetManager
from spaceone.cost_analysis.manager.data_source_manager import DataSourceManager
from spaceone.cost_analysis.model.budget_usage_model import BudgetUsage
from spaceone.cost_analysis.model.budget_model import Budget

_LOGGER = logging.getLogger(__name__)


class BudgetUsageManager(BaseManager):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.budget_mgr: BudgetManager = self.locator.get_manager('BudgetManager')
        self.budget_usage_model: BudgetUsage = self.locator.get_model('BudgetUsage')
        self.identity_mgr: IdentityManager = self.locator.get_manager('IdentityManager')
        self.notification_mgr: NotificationManager = self.locator.get_manager('NotificationManager')
        self.data_source_mgr: DataSourceManager = self.locator.get_manager('DataSourceManager')

    def create_budget_usages(self, budget_vo: Budget):
        if budget_vo.time_unit == 'TOTAL':
            start_dt = datetime.strptime(budget_vo.start, '%Y-%m')
            end_dt = datetime.strptime(budget_vo.end, '%Y-%m')

            dts = [dt for dt in rrule(MONTHLY, dtstart=start_dt, until=end_dt)]
            limit_per_month = round(budget_vo.limit / len(dts), 3)

            for dt in dts:
                budget_usage_data = {
                    'budget_id': budget_vo.budget_id,
                    'name': budget_vo.name,
                    'date': dt.strftime("%Y-%m"),
                    'cost': 0,
                    'limit': limit_per_month,
                    'currency': budget_vo.currency,
                    'provider_filter': budget_vo.provider_filter.to_dict(),
                    'budget': budget_vo,
                    'project_id': budget_vo.project_id,
                    'project_group_id': budget_vo.project_group_id,
                    'data_source_id': budget_vo.data_source_id,
                    'domain_id': budget_vo.domain_id
                }

                self.budget_usage_model.create(budget_usage_data)

        else:
            for planned_limit in budget_vo.planned_limits:
                budget_usage_data = {
                    'budget_id': budget_vo.budget_id,
                    'name': budget_vo.name,
                    'date': planned_limit['date'],
                    'cost': 0,
                    'limit': planned_limit.limit,
                    'currency': budget_vo.currency,
                    'provider_filter': budget_vo.provider_filter.to_dict(),
                    'budget': budget_vo,
                    'project_id': budget_vo.project_id,
                    'project_group_id': budget_vo.project_group_id,
                    'data_source_id': budget_vo.data_source_id,
                    'domain_id': budget_vo.domain_id
                }

                self.budget_usage_model.create(budget_usage_data)

    def update_budget_usage_by_vo(self, params, budget_usage_vo):
        def _rollback(old_data):
            _LOGGER.info(f'[update_budget_usage_by_vo._rollback] Revert Data : '
                         f'{old_data["budget_id"]} / {old_data["date"]}')
            budget_usage_vo.update(old_data)

        self.transaction.add_rollback(_rollback, budget_usage_vo.to_dict())
        return budget_usage_vo.update(params)

    def update_cost_usage(self, budget_id, domain_id):
        _LOGGER.info(f'[update_cost_usage] Update Budget Usage: {budget_id}')
        cost_mgr: CostManager = self.locator.get_manager('CostManager')

        budget_vo = self.budget_mgr.get_budget(budget_id, domain_id)
        self._update_monthly_budget_usage(budget_vo, cost_mgr)

    def update_budget_usage(self, domain_id, data_source_id):
        budget_vos = self.budget_mgr.filter_budgets(domain_id=domain_id, data_source_id=data_source_id)
        for budget_vo in budget_vos:
            self.update_cost_usage(budget_vo.budget_id, domain_id)
            self.notify_budget_usage(budget_vo)

    def notify_budget_usage(self, budget_vo: Budget):
        budget_id = budget_vo.budget_id
        domain_id = budget_vo.domain_id
        current_month = datetime.now().strftime('%Y-%m')
        updated_notifications = []
        is_changed = False
        for notification in budget_vo.notifications:
            if current_month not in notification.notified_months:
                unit = notification.unit
                threshold = notification.threshold
                notification_type = notification.notification_type
                is_notify = False

                if budget_vo.time_unit == 'MONTHLY':
                    budget_usage_vos = self.filter_budget_usages(budget_id=budget_id, domain_id=domain_id)
                    total_budget_usage = sum([budget_usage_vo.cost for budget_usage_vo in budget_usage_vos])
                    budget_limit = budget_vo.limit
                else:
                    budget_usage_vos = self.filter_budget_usages(budget_id=budget_id, domain_id=domain_id,
                                                                 date=current_month)
                    total_budget_usage = budget_usage_vos[0].cost
                    budget_limit = budget_usage_vos[0].limit

                if budget_limit == 0:
                    _LOGGER.debug(f'[notify_budget_usage] budget_limit is 0: {budget_id}')
                    continue

                budget_percentage = round(total_budget_usage / budget_limit * 100, 2)

                if unit == 'PERCENT':
                    if budget_percentage > threshold:
                        is_notify = True
                        is_changed = True
                else:
                    if total_budget_usage > threshold:
                        is_notify = True
                        is_changed = True

                if is_notify:
                    _LOGGER.debug(f'[notify_budget_usage] notify event: {budget_id} (level: {notification_type})')
                    self._notify_message(budget_vo, current_month, total_budget_usage, budget_limit,
                                         budget_percentage, threshold, unit, notification_type)

                    updated_notifications.append({
                        'threshold': threshold,
                        'unit': unit,
                        'notification_type': notification_type,
                        'notified_months': notification.notified_months + [current_month]
                    })
                else:
                    if unit == 'PERCENT':
                        _LOGGER.debug(f'[notify_budget_usage] skip notification: {budget_id} '
                                      f'(usage percent: {budget_percentage}%, threshold: {threshold}%)')
                    else:
                        _LOGGER.debug(f'[notify_budget_usage] skip notification: {budget_id} '
                                      f'(usage cost: {total_budget_usage}, threshold: {threshold})')

                    updated_notifications.append(notification.to_dict())

            else:
                updated_notifications.append(notification.to_dict())

        if is_changed:
            budget_vo.update({'notifications': updated_notifications})

    def _notify_message(self, budget_vo: Budget, current_month, total_budget_usage, budget_limit, budget_percentage,
                        threshold, unit, notification_type):
        data_source_name = self.data_source_mgr.get_data_source(budget_vo.data_source_id, budget_vo.domain_id).name
        project_name = self.identity_mgr.get_project_name(budget_vo.project_id, budget_vo.domain_id)

        if unit == 'PERCENT':
            threshold_str = f'{int(threshold)}%'
        else:
            threshold_str = format(int(threshold), ',')

        description = f'Please check the budget usage and increase the budget limit if necessary.\n\n'
        description += (f'Budget Usage (Currency: {budget_vo.currency}): \n'
                        f'- Usage Cost: {format(round(total_budget_usage, 2), ",")}\n'
                        f'- Limit: {format(budget_limit, ",")}\n'
                        f'- Percentage: {budget_percentage}%\n'
                        f'- Threshold: > {threshold_str}\n')

        if budget_vo.time_unit == 'MONTHLY':
            period = f'{current_month} ~ {current_month}'
        else:
            period = f'{budget_vo.start} ~ {budget_vo.end}'

        message = {
            'resource_type': 'identity.Project',
            'resource_id': budget_vo.project_id,
            'notification_type': 'WARNING' if notification_type == 'WARNING' else 'ERROR',
            'topic': 'cost_analysis.Budget',
            'message': {
                'title': f'Budget usage exceeded - {budget_vo.name}',
                'description': description,
                'tags': [
                    {
                        'key': 'Budget ID',
                        'value': budget_vo.budget_id,
                        'options': {
                            'short': True
                        }
                    },
                    {
                        'key': 'Budget Name',
                        'value': budget_vo.name,
                        'options': {
                            'short': True
                        }
                    },
                    {
                        'key': 'Data Source',
                        'value': data_source_name,
                        'options': {
                            'short': True
                        }
                    },
                    {
                        'key': 'Period',
                        'value': period,
                        'options': {
                            'short': True
                        }
                    },
                    {
                        'key': 'Project',
                        'value': project_name,
                    }
                ],
                'occurred_at': utils.datetime_to_iso8601(datetime.utcnow()),
            },
            'notification_level': 'ALL',
            'domain_id': budget_vo.domain_id
        }

        self.notification_mgr.create_notification(message)

    def filter_budget_usages(self, **conditions):
        return self.budget_usage_model.filter(**conditions)

    def list_budget_usages(self, query={}):
        return self.budget_usage_model.query(**query)

    def stat_budget_usages(self, query):
        return self.budget_usage_model.stat(**query)

    def analyze_budget_usages(self, query):
        query['date_field'] = 'date'
        query['date_field_format'] = '%Y-%m'
        return self.budget_usage_model.analyze(**query)

    def _update_monthly_budget_usage(self, budget_vo: Budget, cost_mgr: CostManager):
        update_data = {}
        query = self._make_cost_analyze_query(budget_vo)
        _LOGGER.debug(f'[_update_monthly_budget_usage]: query: {query}')

        result = cost_mgr.analyze_costs_by_granularity(query, budget_vo.domain_id, budget_vo.data_source_id)
        for cost_usage_data in result.get('results', []):
            if date := cost_usage_data.get('date'):
                update_data[date] = cost_usage_data.get('cost', 0)

        budget_usage_vos = self.budget_usage_model.filter(budget_id=budget_vo.budget_id)
        for budget_usage_vo in budget_usage_vos:
            if budget_usage_vo.date in update_data:
                budget_usage_vo.update({'cost': update_data[budget_usage_vo.date]})
            else:
                budget_usage_vo.update({'cost': 0})

    def _make_cost_analyze_query(self, budget_vo: Budget):
        query = {
            'granularity': 'MONTHLY',
            'start': budget_vo.start,
            'end': budget_vo.end,
            'fields': {
                'cost': {
                    'key': 'cost',
                    'operator': 'sum'
                }
            },
            'filter': [
                {'k': 'domain_id', 'v': budget_vo.domain_id, 'o': 'eq'},
                {'k': 'data_source_id', 'v': budget_vo.data_source_id, 'o': 'eq'},
            ]
        }

        if budget_vo.project_id:
            query['filter'].append({'k': 'project_id', 'v': budget_vo.project_id, 'o': 'eq'})

        else:
            identity_mgr: IdentityManager = self.locator.get_manager('IdentityManager')
            response = identity_mgr.list_projects_in_project_group(budget_vo.project_group_id,
                                                                   budget_vo.domain_id, True)

            project_ids = []
            for project_info in response.get('results', []):
                project_ids.append(project_info['project_id'])

            query['filter'].append({'k': 'project_id', 'v': project_ids, 'o': 'in'})

        if budget_vo.provider_filter and budget_vo.provider_filter.state == 'ENABLED':
            query['filter'].append({'k': 'provider', 'v': budget_vo.provider_filter.providers, 'o': 'in'})

        return query



from mongoengine import *

from spaceone.core.model.mongo_model import MongoModel


class Cost(MongoModel):
    cost_id = StringField(max_length=40, generate_id='cost')
    cost = FloatField(required=True)
    usage_quantity = FloatField(default=0)
    usage_unit = StringField(max_length=255, default=None, null=True)
    provider = StringField(max_length=40, default=None, null=True)
    region_code = StringField(max_length=255, default=None, null=True)
    region_key = StringField(max_length=255, default=None, null=True)
    product = StringField(max_length=255, default=None, null=True)
    usage_type = StringField(max_length=255, default=None, null=True)
    resource = StringField(default=None, null=True)
    tags = DictField(default={})
    additional_info = DictField(default={})
    service_account_id = StringField(max_length=40, default=None, null=True)
    project_id = StringField(max_length=40, default=None, null=True)
    project_group_id = StringField(max_length=40, default=None, null=True)
    data_source_id = StringField(max_length=40)
    job_id = StringField(max_length=40, default=None, null=True)
    job_task_id = StringField(max_length=40, default=None, null=True)
    domain_id = StringField(max_length=40)
    billed_year = StringField(max_length=4, required=True)
    billed_month = StringField(max_length=7, required=True)
    billed_date = StringField(max_length=10, required=True)

    meta = {
        'updatable_fields': [],
        'minimal_fields': [
            'cost_id',
            'cost',
            'provider',
            'region_code',
            'product',
            'usage_type',
            'resource',
            'data_source_id',
            'billed_date'
        ],
        'change_query_keys': {
            'user_projects': 'project_id'
        },
        'indexes': [
            {
                "fields": ['domain_id', 'data_source_id', 'job_id', 'job_task_id', '-billed_date'],
                "name": "COMPOUND_INDEX_FOR_SYNC_JOB_1"
            },
            {
                "fields": ['domain_id', 'data_source_id', 'job_id', '-billed_month'],
                "name": "COMPOUND_INDEX_FOR_SYNC_JOB_2"
            },
            {
                "fields": ['domain_id', 'data_source_id', '-billed_date', 'project_id', 'cost'],
                "name": "COMPOUND_INDEX_FOR_SEARCH"
            },
            {
                "fields": ['domain_id', 'cost_id', 'project_id'],
                "name": "COMPOUND_INDEX_FOR_DELETE"
            }
        ]
    }


class MonthlyCost(MongoModel):
    cost = FloatField(required=True)
    usage_quantity = FloatField(default=0)
    usage_unit = StringField(max_length=255, default=None, null=True)
    provider = StringField(max_length=40, default=None, null=True)
    region_code = StringField(max_length=255, default=None, null=True)
    region_key = StringField(max_length=255, default=None, null=True)
    product = StringField(max_length=255, default=None, null=True)
    usage_type = StringField(max_length=255, default=None, null=True)
    resource = StringField(default=None, null=True)
    tags = DictField(default={})
    additional_info = DictField(default={})
    service_account_id = StringField(max_length=40, default=None, null=True)
    project_id = StringField(max_length=40, default=None, null=True)
    project_group_id = StringField(max_length=40, default=None, null=True)
    data_source_id = StringField(max_length=40)
    job_id = StringField(max_length=40, default=None, null=True)
    job_task_id = StringField(max_length=40, default=None, null=True)
    domain_id = StringField(max_length=40)
    billed_year = StringField(max_length=4, required=True)
    billed_month = StringField(max_length=7, required=True)

    meta = {
        'updatable_fields': [],
        'change_query_keys': {
            'user_projects': 'project_id'
        },
        'indexes': [
            {
                "fields": ['domain_id', 'data_source_id', 'job_id', '-billed_month'],
                "name": "COMPOUND_INDEX_FOR_SYNC_JOB"
            },
            {
                "fields": ['domain_id', 'data_source_id', '-billed_month', 'project_id', 'cost'],
                "name": "COMPOUND_INDEX_FOR_SEARCH_1"
            },
            {
                "fields": ['domain_id', 'data_source_id', '-billed_year', 'project_id', 'cost'],
                "name": "COMPOUND_INDEX_FOR_SEARCH_2"
            },
        ]
    }


class CostQueryHistory(MongoModel):
    query_hash = StringField(max_length=255)
    query_options = DictField(default={})
    data_source_id = StringField(max_length=40)
    domain_id = StringField(max_length=40)
    updated_at = DateTimeField(auto_now=True)

    meta = {
        'updatable_fields': [
            'updated_at'
        ],
        'indexes': [
            {
                "fields": ['domain_id', 'data_source_id', 'query_hash'],
                "name": "COMPOUND_INDEX_FOR_SEARCH"
            },
        ]
    }
